(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;

/* Package-scope variables */
var SandstormDb, Packages, DevApps, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, isDemoUser, isSignedUp, isSignedUpOrDemo, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, roleAssignmentPattern, addStaticAsset, userPictureUrl, fetchPicture;

(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/db.js                                                                        //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
// This file defines the database schema.                                                             // 17
                                                                                                      // 18
// Useful for debugging: Set the env variable LOG_MONGO_QUERIES to have the server write every        // 19
// query it makes, so you can see if it's doing queries too often, etc.                               // 20
if (Meteor.isServer && process.env.LOG_MONGO_QUERIES) {                                               // 21
  var oldFind = Mongo.Collection.prototype.find;                                                      // 22
  Mongo.Collection.prototype.find = function () {                                                     // 23
    console.log(this._prefix, arguments);                                                             // 24
    return oldFind.apply(this, arguments);                                                            // 25
  }                                                                                                   // 26
}                                                                                                     // 27
                                                                                                      // 28
// Users = new Mongo.Collection("users");                                                             // 29
// The users collection is special and can be accessed through `Meteor.users`.                        // 30
// See https://docs.meteor.com/#/full/meteor_users. Entries in the users collection correspond to     // 31
// accounts. Each represents a capability store belonging to a human.                                 // 32
//                                                                                                    // 33
// Note that accounts are not the same thing as identities. An account may have multiple identities   // 34
// attached to it. Identities provide a way to authenticate the user and are also used to present     // 35
// globally unique and stable user indentifiers to grains via `Grain.UserInfo.userId`.                // 36
//                                                                                                    // 37
// Each entry in this collection contains:                                                            // 38
//   _id: Random.                                                                                     // 39
//   createdAt: Date when this entry was added to the collection.                                     // 40
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.               // 41
//   profile: Object containing profile data editable by the user and visible to other users. May     // 42
//            include the following fields. Note that if any field is missing, the first fallback     // 43
//            is to check `services` for details provided by the identity provider (the details       // 44
//            of which differ per-provider). Only if that is also missing do we fall back to          // 45
//            defaults.                                                                               // 46
//       name: String containing the display name of the user. Default: first part of email.          // 47
//       handle: String containing the user's preferred handle. Default: first part of email.         // 48
//       picture: _id into the StaticAssets table for the user's picture. Default: identicon.         // 49
//       pronoun: One of "male", "female", "neutral", or "robot". Default: neutral.                   // 50
//   services: Object containing login and identity data used by Meteor authentication services.      // 51
//   isAdmin: Boolean indicating whether this user is allowed to access the Sandstorm admin panel.    // 52
//   signupKey: If this is an invited user, then this field contains their signup key.                // 53
//   signupNote: If the user was invited through a link, then this field contains the note that the   // 54
//               inviter admin attached to the key.                                                   // 55
//   signupEmail: If the user was invited by email, then this field contains the email address that   // 56
//                the invite was sent to.                                                             // 57
//   plan: _id of an entry in the Plans table which determines the user's qutoa.                      // 58
//   storageUsage: Number of bytes this user is currently storing.                                    // 59
//   expires: Date when this user's account should be deleted. Only present for demo users.           // 60
//   isAppDemoUser: True if this is a demo user who arrived via an /appdemo/ link.                    // 61
//   appDemoId: If this is an appdemo user (see above), the app ID they started out demoing.          // 62
//   payments: Object defined by payments module, if loaded.                                          // 63
//   dailySentMailCount: Number of emails sent by this user today; used to limit spam.                // 64
                                                                                                      // 65
Packages = new Mongo.Collection("packages");                                                          // 66
// Packages which are installed or downloading.                                                       // 67
//                                                                                                    // 68
// Each contains:                                                                                     // 69
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                   // 70
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"  // 71
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),               // 72
//       >1 = download byte count.                                                                    // 73
//   error:  If status is "failed", error message string.                                             // 74
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.         // 75
//   appId:  If status is "ready", the application ID string.  Packages representing different        // 76
//       versions of the same app have the same appId.  The spk tool defines the app ID format        // 77
//       and can cryptographically verify that a package belongs to a particular app ID.              // 78
//   shouldCleanup:  If true, a reference to this package was recently dropped, and the package       // 79
//       collector should at some point check whether there are any other references and, if not,     // 80
//       delete the package.                                                                          // 81
//   url:  When status is "download", the URL from which the SPK can be obtained, if provided.        // 82
                                                                                                      // 83
DevApps = new Mongo.Collection("devapps");                                                            // 84
// List of applications currently made available via the dev tools running on the local machine.      // 85
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool    // 86
// on the local machine to publish an under-development app to this server. That should only ever     // 87
// happen on developers' desktop machines.                                                            // 88
//                                                                                                    // 89
// While a dev app is published, it automatically appears as installed by every user of the server,   // 90
// and it overrides all packages with the same application ID. If any instances of those packages     // 91
// are currently open, they are killed and reset on publish.                                          // 92
//                                                                                                    // 93
// When the dev tool disconnects, the app is automatically unpublished, and any open instances        // 94
// are again killed and refreshed.                                                                    // 95
//                                                                                                    // 96
// Each contains:                                                                                     // 97
//   _id:  The application ID string (as with Packages.appId).                                        // 98
//   packageId:  The directory name where the dev package is mounted.                                 // 99
//   timestamp:  Time when the package was last updated. If this changes while the package is         // 100
//     published, all running instances are reset. This is used e.g. to reset the app each time       // 101
//     changes are made to the source code.                                                           // 102
//   manifest:  The app's manifest, as with Packages.manifest.                                        // 103
                                                                                                      // 104
UserActions = new Mongo.Collection("userActions");                                                    // 105
// List of actions that each user has installed which create new grains.  Each app may install        // 106
// some number of actions (usually, one).                                                             // 107
//                                                                                                    // 108
// Each contains:                                                                                     // 109
//   _id:  random                                                                                     // 110
//   userId:  User who has installed this action.                                                     // 111
//   packageId:  Package used to run this action.                                                     // 112
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.               // 113
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle; denormalized so                // 114
//       that clients can access it without subscribing to the Packages collection.                   // 115
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for           // 116
//       searchability.                                                                               // 117
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"              // 118
//   title: JSON-encoded LocalizedText title for this action, e.g.                                    // 119
//       `{defaultText: "New Spreadsheet"}`.                                                          // 120
//   nounPhrase: JSON-encoded LocalizedText describing what is created when this action is run.       // 121
//   command:  Manifest.Command to run this action (see package.capnp).                               // 122
                                                                                                      // 123
Grains = new Mongo.Collection("grains");                                                              // 124
// Grains belonging to users.                                                                         // 125
//                                                                                                    // 126
// Each contains:                                                                                     // 127
//   _id:  random                                                                                     // 128
//   packageId:  _id of the package of which this grain is an instance.                               // 129
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.               // 130
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for           // 131
//       searchability.                                                                               // 132
//   userId:  User who owns this grain.                                                               // 133
//   title:  Human-readable string title, as chosen by the user.                                      // 134
//   lastUsed:  Date when the grain was last used by a user.                                          // 135
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                   // 136
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent   // 137
//                   time a session to this grain was opened.                                         // 138
//                                                                                                    // 139
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and    // 140
// web publishing functionality without powerbox support; they will be replaced once the powerbox     // 141
// is implemented.                                                                                    // 142
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and    // 143
//       web publishing. This field is initialized when first requested by the app.                   // 144
                                                                                                      // 145
RoleAssignments = new Mongo.Collection("roleAssignments");                                            // 146
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges   // 147
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens            // 148
// collection.                                                                                        // 149
                                                                                                      // 150
Contacts = new Mongo.Collection("contacts");                                                          // 151
// Edges in the social graph.                                                                         // 152
//                                                                                                    // 153
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob        // 154
// will show up in her user-picker UI for actions like share-by-identity.                             // 155
//                                                                                                    // 156
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of       // 157
// Bob's.                                                                                             // 158
//                                                                                                    // 159
// Each contains:                                                                                     // 160
//   _id: random                                                                                      // 161
//   ownerId: The `_id` of the user who owns this contact.                                            // 162
//   userId:  The `_id` of the contacted user.                                                        // 163
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies       // 164
//            the contact to the owner.                                                               // 165
//   created: Date when this contact was created.                                                     // 166
                                                                                                      // 167
Sessions = new Mongo.Collection("sessions");                                                          // 168
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.   // 169
//                                                                                                    // 170
// Each contains:                                                                                     // 171
//   _id:  random                                                                                     // 172
//   grainId:  _id of the grain to which this session is connected.                                   // 173
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the    // 174
//       '*' in WILDCARD_HOST.                                                                        // 175
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some       // 176
//       period.                                                                                      // 177
//   userId:  User who owns this session.                                                             // 178
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens     // 179
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),        // 180
//       anonymous users can get access without an API token and so neither userId nor hashedToken    // 181
//       are present.                                                                                 // 182
//   powerboxView: If present, this is a view that should be presented as part of a powerbox          // 183
//       interaction.                                                                                 // 184
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                  // 185
//   viewInfo: The UiView.ViewInfo corresponding to the underlying UiSession. This isn't populated    // 186
//       until newSession is called on the UiView.                                                    // 187
//   permissions: The permissions for the current user on this UiView. This isn't populated           // 188
//       until newSession is called on the UiView.                                                    // 189
//   hasLoaded: Marked as true by the proxy when the underlying UiSession has responded to its first  // 190
//       request                                                                                      // 191
                                                                                                      // 192
SignupKeys = new Mongo.Collection("signupKeys");                                                      // 193
// Invite keys which may be used by users to get access to Sandstorm.                                 // 194
//                                                                                                    // 195
// Each contains:                                                                                     // 196
//   _id:  random                                                                                     // 197
//   used:  Boolean indicating whether this key has already been consumed.                            // 198
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.         // 199
//   email: If this key was sent as an email invite, the email address to which it was sent.          // 200
                                                                                                      // 201
ActivityStats = new Mongo.Collection("activityStats");                                                // 202
// Contains usage statistics taken on a regular interval. Each entry is a data point.                 // 203
//                                                                                                    // 204
// Each contains:                                                                                     // 205
//   timestamp: Date when measurements were taken.                                                    // 206
//   daily: Contains stats counts pertaining to the last day before the sample time.                  // 207
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.          // 208
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.          // 209
//                                                                                                    // 210
// Each of daily, weekly, and monthly contains:                                                       // 211
//   activeUsers: The number of unique users who have used a grain on the server in the time          // 212
//       interval. Only counts logged-in users.                                                       // 213
//   demoUsers: Demo users.                                                                           // 214
//   appDemoUsers: Users that came in through "app demo".                                             // 215
//   activeGrains: The number of unique grains that have been used in the time interval.              // 216
//   apps: An object indexed by app ID recording, for each app:                                       // 217
//       owners: Number of unique owners of this app (counting only grains that still exist).         // 218
//       sharedUsers: Number of users who have accessed other people's grains of this app (counting   // 219
//         only grains that still exist).                                                             // 220
//       grains: Number of active grains of this app (that still exist).                              // 221
//       deleted: Number of non-demo grains of this app that were deleted.                            // 222
//       demoed: Number of demo grains created and expired.                                           // 223
//       appDemoUsers: Number of app demos initiated with this app.                                   // 224
                                                                                                      // 225
DeleteStats = new Mongo.Collection("deleteStats");                                                    // 226
// Contains records of objects that were deleted, for stat-keeping purposes.                          // 227
//                                                                                                    // 228
// Each contains:                                                                                     // 229
//   type: "grain" or "user" or "demoGrain" or "demoUser" or "appDemoUser"                            // 230
//   lastActive: Date of the user's or grain's last activity.                                         // 231
//   appId: For type = "grain", the app ID of the grain. For type = "appDemoUser", the app ID they    // 232
//     arrived to demo. For others, undefined.                                                        // 233
                                                                                                      // 234
FileTokens = new Mongo.Collection("fileTokens");                                                      // 235
// Tokens corresponding to backup files that are currently stored on the server. A user receives      // 236
// a token when they create a backup file (either by uploading it, or by backing up one of their      // 237
// grains) and may use the token to read the file (either to download it, or to restore a new         // 238
// grain from it).                                                                                    // 239
//                                                                                                    // 240
// Each contains:                                                                                     // 241
//   _id:       The unguessable token string.                                                         // 242
//   name:      Suggested filename.                                                                   // 243
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.       // 244
                                                                                                      // 245
ApiTokens = new Mongo.Collection("apiTokens");                                                        // 246
// Access tokens for APIs exported by apps.                                                           // 247
//                                                                                                    // 248
// Originally API tokens were only used by external users through the HTTP API endpoint. However,     // 249
// now they are also used to implement SturdyRefs, not just held by external users, but also when     // 250
// an app holds a SturdyRef to another app within the same server. See the various `save()`,          // 251
// `restore()`, and `drop()` methods in `grain.capnp` (on `SansdtormApi`, `AppPersistent`, and        // 252
// `MainView`) -- the fields of type `Data` are API tokens.                                           // 253
//                                                                                                    // 254
// Each contains:                                                                                     // 255
//   _id:       A SHA-256 hash of the token.                                                          // 256
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)     // 257
//   userId:    For UiView capabilities, this is the user for whom the view is attenuated. That       // 258
//              is, the UiView's newSession() method will intersect the requested permissions with    // 259
//              this user's permissions before forwarding on to the underlying app. If `userId` is    // 260
//              not present, then no user attenuation is applied, i.e. this is a raw UiView as        // 261
//              implemented by the app. (The `roleAssignment` field, below, may still apply.)         // 262
//              For non-UiView capabilities, `userId` is never present.                               // 263
//              Note that this is NOT the user against whom the `requiredPermissions` parameter of    // 264
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerUser`.      // 265
//   userInfo:  *DEPRECATED* For API tokens created by the app through HackSessionContext, the        // 266
//              UserInfo struct that should be passed to `newSession()` when exercising this token,   // 267
//              in decoded (JS object) format. This is a temporary hack. `userId` is never present    // 268
//              when `userInfo` is present.                                                           // 269
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded        // 270
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These   // 271
//              permissions will be intersected with those held by `userId` when the view is opened.  // 272
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as   // 273
//              anonymous rather than as directly associated with `userId`. This has no effect on     // 274
//              the permissions granted.                                                              // 275
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by     // 276
//              the app or its supervisor (whereas without this it strictly represents UiView).       // 277
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).     // 278
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is            // 279
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.              // 280
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,   // 281
//              not a particular grain. (`grainId` and `userId` are not set.) This is an object       // 282
//              containing exactly one of the following fields:                                       // 283
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                   // 284
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the          // 285
//                           `Notifications` collection.                                              // 286
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this     // 287
//                  will be moved out of the frontend and into the backend, but we'll migrate the     // 288
//                  database when that happens. This field contains the boolean true to signify that  // 289
//                  it has been set.                                                                  // 290
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                    // 291
//   parentToken: If present, then this token represents exactly the capability represented by        // 292
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated   // 293
//              by `roleAssignment` (if present). To facilitate permissions computations, if the      // 294
//              capability is a UiView, then `grainId` is set to the backing grain and `userId` is    // 295
//              set to the user who shared the view. None of `userInfo`, `objectId`, or               // 296
//              `frontendRef` are present when `parentToken` is present.                              // 297
//   petname:   Human-readable label for this access token, useful for identifying tokens for         // 298
//              revocation. This should be displayed when visualizing incoming capabilities to        // 299
//              the grain identified by `grainId`.                                                    // 300
//   created:   Date when this token was created.                                                     // 301
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may   // 302
//              become un-revoked in the future.                                                      // 303
//   expires:   Optional expiration Date. If undefined, the token does not expire.                    // 304
//   owner:     A `ApiTokenRefOwner` (defined in `supervisor.capnp`, stored as a JSON object)         // 305
//              as passed to the `save()` call that created this token. If not present, treat         // 306
//              as `webkey` (the default for `ApiTokenOwner`).                                        // 307
//   expiresIfUnused:                                                                                 // 308
//              Optional Date after which the token, if it has not been used yet, expires.            // 309
//              This field should be cleared on a token's first use.                                  // 310
//   requirements: List of conditions which must hold for this token to be considered valid.          // 311
//              Semantically, this list specifies the powers which were *used* to originally          // 312
//              create the token. If any condition in the list becomes untrue, then the token must    // 313
//              be considered revoked, and all live refs and sturdy refs obtained transitively        // 314
//              through it must also become revoked. Each item is the JSON serialization of the       // 315
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                        // 316
//                                                                                                    // 317
// It is important to note that a token's owner and provider are independent from each other. To      // 318
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:   // 319
//                                                                                                    // 320
// struct ApiToken {                                                                                  // 321
//   owner :ApiTokenOwner;                                                                            // 322
//   provider :union {                                                                                // 323
//     grain :group {                                                                                 // 324
//       grainId :Text;                                                                               // 325
//       union {                                                                                      // 326
//         uiView :group {                                                                            // 327
//           userId :Text;                                                                            // 328
//           roleAssignment :RoleAssignment;                                                          // 329
//           forSharing :Bool;                                                                        // 330
//         }                                                                                          // 331
//         objectId :SupervisorObjectId;                                                              // 332
//       }                                                                                            // 333
//     }                                                                                              // 334
//     frontendRef :union {                                                                           // 335
//        notificationHandle :Text;                                                                   // 336
//        ipNetwork :Bool;                                                                            // 337
//        ipInterface :Bool;                                                                          // 338
//     }                                                                                              // 339
//     child :group {                                                                                 // 340
//       parentToken :Text;                                                                           // 341
//       union {                                                                                      // 342
//         uiView :group {                                                                            // 343
//           grainId :Text;                                                                           // 344
//           userId :Text;                                                                            // 345
//           roleAssignment :RoleAssignment = (allAccess = ());                                       // 346
//         }                                                                                          // 347
//         other :Void;                                                                               // 348
//       }                                                                                            // 349
//     }                                                                                              // 350
//   }                                                                                                // 351
//   requirements: List(Supervisor.MembraneRequirement);                                              // 352
//   ...                                                                                              // 353
// }                                                                                                  // 354
                                                                                                      // 355
Notifications = new Mongo.Collection("notifications");                                                // 356
// Notifications for a user.                                                                          // 357
//                                                                                                    // 358
// Each contains:                                                                                     // 359
//   _id:          random                                                                             // 360
//   ongoing:      If present, this is an ongoing notification, and this field contains an            // 361
//                 ApiToken referencing the `OngoingNotification` capability.                         // 362
//   grainId:      The grain originating this notification, if any.                                   // 363
//   userId:       The user receiving the notification.                                               // 364
//   text:         The JSON-ified LocalizedText to display in the notification.                       // 365
//   isUnread:     Boolean indicating if this notification is unread.                                 // 366
//   timestamp:    Date when this notification was last updated                                       // 367
                                                                                                      // 368
StatsTokens = new Mongo.Collection("statsTokens");                                                    // 369
// Access tokens for the Stats collection                                                             // 370
//                                                                                                    // 371
// These tokens are used for accessing the ActivityStats collection remotely                          // 372
// (ie. from a dashboard webapp)                                                                      // 373
//                                                                                                    // 374
// Each contains:                                                                                     // 375
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                 // 376
                                                                                                      // 377
Misc = new Mongo.Collection("misc");                                                                  // 378
// Miscellaneous configuration and other settings                                                     // 379
//                                                                                                    // 380
// This table is currently only used for persisting BASE_URL from one session to the next,            // 381
// but in general any miscellaneous settings should go in here                                        // 382
//                                                                                                    // 383
// Each contains:                                                                                     // 384
//   _id:       The name of the setting. eg. "BASE_URL"                                               // 385
//   value:     The value of the setting.                                                             // 386
                                                                                                      // 387
Settings = new Mongo.Collection("settings");                                                          // 388
// Settings for this Sandstorm instance go here. They are configured through the adminSettings        // 389
// route. This collection differs from misc in that this collection is completely user controlled.    // 390
//                                                                                                    // 391
// Each contains:                                                                                     // 392
//   _id:       The name of the setting. eg. "MAIL_URL"                                               // 393
//   value:     The value of the setting.                                                             // 394
//   potentially other fields that are unique to the setting                                          // 395
                                                                                                      // 396
Migrations = new Mongo.Collection("migrations");                                                      // 397
// This table tracks which migrations we have applied to this instance.                               // 398
// It contains a single entry:                                                                        // 399
//   _id:       "migrations_applied"                                                                  // 400
//   value:     The number of migrations this instance has successfully completed.                    // 401
                                                                                                      // 402
StaticAssets = new Mongo.Collection("staticAssets");                                                  // 403
// Collection of static assets served up from the Sandstorm server's "static" host. We only           // 404
// support relatively small assets: under 1MB each.                                                   // 405
//                                                                                                    // 406
// Each contains:                                                                                     // 407
//   _id:       Random ID; will be used in the URL.                                                   // 408
//   hash:      A SHA-256 hash of the data, used to de-dupe.                                          // 409
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                             // 410
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                   // 411
//   content:   The asset content (byte buffer).                                                      // 412
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't   // 413
//       have transactions, this needs to bias towards over-counting; a backup GC could be used       // 414
//       to catch leaked assets, although it's probably not a big deal in practice.                   // 415
                                                                                                      // 416
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                        // 417
// Collection of tokens representing a single-use permission to upload an asset, such as a new        // 418
// profile picture.                                                                                   // 419
//                                                                                                    // 420
// Each contains:                                                                                     // 421
//   _id:       Random ID.                                                                            // 422
//   purpose:   Contains one of the following, indicating how the asset is to be used:                // 423
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:         // 424
//           userId: User whose picture shall be replaced.                                            // 425
//           identityId: Which of the user's identities shall be updated.                             // 426
//   expires:   Time when this token will go away if unused.                                          // 427
                                                                                                      // 428
Plans = new Mongo.Collection("plans");                                                                // 429
// Subscription plans, which determine quota.                                                         // 430
//                                                                                                    // 431
// Each contains:                                                                                     // 432
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...               // 433
//   storage: Number of bytes this user is allowed to store.                                          // 434
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                         // 435
//   computeLabel: Label to display to the user describing this plan's compute units.                 // 436
//   grains: Total number of grains this user can create (often `Infinity`).                          // 437
//   price: Price per month in US cents.                                                              // 438
                                                                                                      // 439
if (Meteor.isServer) {                                                                                // 440
  Meteor.publish("credentials", function () {                                                         // 441
    // Data needed for isSignedUp() and isAdmin() to work.                                            // 442
                                                                                                      // 443
    if (this.userId) {                                                                                // 444
      return [                                                                                        // 445
        Meteor.users.find({_id: this.userId},                                                         // 446
            {fields: {signupKey: 1, isAdmin: 1, expires: 1, storageUsage: 1,                          // 447
                      plan: 1, hasCompletedSignup: 1}}),                                              // 448
        Plans.find()                                                                                  // 449
      ];                                                                                              // 450
    } else {                                                                                          // 451
      return [];                                                                                      // 452
    }                                                                                                 // 453
  });                                                                                                 // 454
}                                                                                                     // 455
                                                                                                      // 456
isDemoUser = function() {                                                                             // 457
  // Returns true if this is a demo user.                                                             // 458
                                                                                                      // 459
  var user = Meteor.user();                                                                           // 460
  if (user && user.expires) {                                                                         // 461
    return true;                                                                                      // 462
  } else {                                                                                            // 463
    return false;                                                                                     // 464
  }                                                                                                   // 465
}                                                                                                     // 466
                                                                                                      // 467
isSignedUp = function() {                                                                             // 468
  // Returns true if the user has presented an invite key.                                            // 469
                                                                                                      // 470
  var user = Meteor.user();                                                                           // 471
                                                                                                      // 472
  if (!user) return false;  // not signed in                                                          // 473
                                                                                                      // 474
  if (user.expires) return false;  // demo user.                                                      // 475
                                                                                                      // 476
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                    // 477
                                                                                                      // 478
  if (user.signupKey) return true;  // user is invited                                                // 479
                                                                                                      // 480
  return false;                                                                                       // 481
}                                                                                                     // 482
                                                                                                      // 483
isSignedUpOrDemo = function () {                                                                      // 484
  var user = Meteor.user();                                                                           // 485
                                                                                                      // 486
  if (!user) return false;  // not signed in                                                          // 487
                                                                                                      // 488
  if (user.expires) return true;  // demo user.                                                       // 489
                                                                                                      // 490
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                    // 491
                                                                                                      // 492
  if (user.signupKey) return true;  // user is invited                                                // 493
                                                                                                      // 494
  return false;                                                                                       // 495
}                                                                                                     // 496
                                                                                                      // 497
isUserOverQuota = function (user) {                                                                   // 498
  // Return false if user has quota space remaining, true if it is full. When this returns true,      // 499
  // we will not allow the user to create new grains, though they may be able to open existing ones   // 500
  // which may still increase their storage usage.                                                    // 501
  //                                                                                                  // 502
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                 // 503
                                                                                                      // 504
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                             // 505
                                                                                                      // 506
  var plan = Plans.findOne(user.plan || "free");                                                      // 507
                                                                                                      // 508
  if (plan.grains < Infinity) {                                                                       // 509
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains}).count();            // 510
    if (count >= plan.grains) return "outOfGrains";                                                   // 511
  }                                                                                                   // 512
                                                                                                      // 513
  return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";            // 514
}                                                                                                     // 515
                                                                                                      // 516
isUserExcessivelyOverQuota = function (user) {                                                        // 517
  // Return true if user is so far over quota that we should prevent their existing grains from       // 518
  // running at all.                                                                                  // 519
  //                                                                                                  // 520
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                 // 521
                                                                                                      // 522
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                             // 523
                                                                                                      // 524
  var plan = Plans.findOne(user.plan || "free");                                                      // 525
                                                                                                      // 526
  if (plan.grains < Infinity) {                                                                       // 527
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains * 2}).count();        // 528
    if (count >= plan.grains * 2) return "outOfGrains";                                               // 529
  }                                                                                                   // 530
                                                                                                      // 531
  return plan && user.storageUsage && user.storageUsage >= plan.storage * 1.2 && "outOfStorage";      // 532
}                                                                                                     // 533
                                                                                                      // 534
isAdmin = function() {                                                                                // 535
  // Returns true if the user is the administrator.                                                   // 536
                                                                                                      // 537
  var user = Meteor.user();                                                                           // 538
  if (user && user.isAdmin) {                                                                         // 539
    return true;                                                                                      // 540
  } else {                                                                                            // 541
    return false;                                                                                     // 542
  }                                                                                                   // 543
}                                                                                                     // 544
                                                                                                      // 545
isAdminById = function(id) {                                                                          // 546
  // Returns true if the user's id is the administrator.                                              // 547
                                                                                                      // 548
  var user = Meteor.users.findOne({_id: id}, {fields: {isAdmin: 1}})                                  // 549
  if (user && user.isAdmin) {                                                                         // 550
    return true;                                                                                      // 551
  } else {                                                                                            // 552
    return false;                                                                                     // 553
  }                                                                                                   // 554
}                                                                                                     // 555
                                                                                                      // 556
findAdminUserForToken = function (token) {                                                            // 557
  if (!token.requirements) {                                                                          // 558
    return;                                                                                           // 559
  }                                                                                                   // 560
  var requirements = token.requirements.filter(function (requirement) {                               // 561
    return "userIsAdmin" in requirement;                                                              // 562
  });                                                                                                 // 563
                                                                                                      // 564
  if (requirements.length > 1) {                                                                      // 565
    return;                                                                                           // 566
  }                                                                                                   // 567
  if (requirements.length === 0) {                                                                    // 568
    return;                                                                                           // 569
  }                                                                                                   // 570
  return requirements[0].userIsAdmin;                                                                 // 571
};                                                                                                    // 572
                                                                                                      // 573
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                      // 574
                                                                                                      // 575
if (wildcardHost.length != 2) {                                                                       // 576
  throw new Error("Wildcard host must contain exactly one asterisk.");                                // 577
}                                                                                                     // 578
                                                                                                      // 579
matchWildcardHost = function(host) {                                                                  // 580
  // See if the hostname is a member of our wildcard. If so, extract the ID.                          // 581
                                                                                                      // 582
  var prefix = wildcardHost[0];                                                                       // 583
  var suffix = wildcardHost[1];                                                                       // 584
  if (host.lastIndexOf(prefix, 0) >= 0 &&                                                             // 585
      host.indexOf(suffix, -suffix.length) >= 0 &&                                                    // 586
      host.length >= prefix.length + suffix.length) {                                                 // 587
    var id = host.slice(prefix.length, -suffix.length);                                               // 588
    if (id.match(/^[a-z0-9]*$/)) {                                                                    // 589
      return id;                                                                                      // 590
    }                                                                                                 // 591
  }                                                                                                   // 592
                                                                                                      // 593
  return null;                                                                                        // 594
}                                                                                                     // 595
                                                                                                      // 596
makeWildcardHost = function (id) {                                                                    // 597
  return wildcardHost[0] + id + wildcardHost[1];                                                      // 598
}                                                                                                     // 599
                                                                                                      // 600
if (Meteor.isServer) {                                                                                // 601
  var Url = Npm.require("url");                                                                       // 602
  getWildcardOrigin = function () {                                                                   // 603
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're       // 604
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),     // 605
    // an astrisk is only allowed as the first character and must be followed by a period. So we need // 606
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can    // 607
    // really do. We also add the protocol to the front (again, that's what CSP wants).               // 608
                                                                                                      // 609
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.  // 610
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                          // 611
                                                                                                      // 612
    var dotPos = wildcardHost[1].indexOf(".");                                                        // 613
    if (dotPos < 0) {                                                                                 // 614
      return protocol + "//*";                                                                        // 615
    } else {                                                                                          // 616
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                        // 617
    }                                                                                                 // 618
  }                                                                                                   // 619
}                                                                                                     // 620
                                                                                                      // 621
allowDevAccounts = function () {                                                                      // 622
  var setting = Settings.findOne({_id: "devAccounts"});                                               // 623
  if (setting) {                                                                                      // 624
    return setting.value;                                                                             // 625
  } else {                                                                                            // 626
    return Meteor.settings && Meteor.settings.public &&                                               // 627
           Meteor.settings.public.allowDevAccounts;                                                   // 628
  }                                                                                                   // 629
};                                                                                                    // 630
                                                                                                      // 631
roleAssignmentPattern = {                                                                             // 632
  none : Match.Optional(null),                                                                        // 633
  allAccess: Match.Optional(null),                                                                    // 634
  roleId: Match.Optional(Match.Integer),                                                              // 635
  addPermissions: Match.Optional([Boolean]),                                                          // 636
  removePermissions: Match.Optional([Boolean]),                                                       // 637
};                                                                                                    // 638
                                                                                                      // 639
SandstormDb = function () {                                                                           // 640
  this.collections = {                                                                                // 641
    // Direct access to underlying collections. DEPRECATED.                                           // 642
    //                                                                                                // 643
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove     // 644
    //   direct access to the collections.                                                            // 645
                                                                                                      // 646
    packages: Packages,                                                                               // 647
    devApps: DevApps,                                                                                 // 648
    userActions: UserActions,                                                                         // 649
    grains: Grains,                                                                                   // 650
    contacts: Contacts,                                                                               // 651
    sessions: Sessions,                                                                               // 652
    signupKeys: SignupKeys,                                                                           // 653
    activityStats: ActivityStats,                                                                     // 654
    deleteStats: DeleteStats,                                                                         // 655
    fileTokens: FileTokens,                                                                           // 656
    apiTokens: ApiTokens,                                                                             // 657
    notifications: Notifications,                                                                     // 658
    statsTokens: StatsTokens,                                                                         // 659
    misc: Misc,                                                                                       // 660
    settings: Settings,                                                                               // 661
                                                                                                      // 662
    // Intentionally omitted:                                                                         // 663
    // - Migrations, since it's used only within this package.                                        // 664
    // - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.   // 665
  };                                                                                                  // 666
};                                                                                                    // 667
                                                                                                      // 668
// TODO(cleanup): These methods should not be defined freestanding and should use collection          // 669
//   objects created in SandstormDb's constructor rather than globals.                                // 670
                                                                                                      // 671
_.extend(SandstormDb.prototype, {                                                                     // 672
  isDemoUser: isDemoUser,                                                                             // 673
  isSignedUp: isSignedUp,                                                                             // 674
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                 // 675
  isUserOverQuota: isUserOverQuota,                                                                   // 676
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                             // 677
  isAdmin: isAdmin,                                                                                   // 678
  isAdminById: isAdminById,                                                                           // 679
  findAdminUserForToken: findAdminUserForToken,                                                       // 680
  matchWildcardHost: matchWildcardHost,                                                               // 681
  makeWildcardHost: makeWildcardHost,                                                                 // 682
  allowDevAccounts: allowDevAccounts,                                                                 // 683
  roleAssignmentPattern: roleAssignmentPattern,                                                       // 684
});                                                                                                   // 685
                                                                                                      // 686
if (Meteor.isServer) {                                                                                // 687
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                        // 688
}                                                                                                     // 689
                                                                                                      // 690
// =======================================================================================            // 691
// Below this point are newly-written or refactored functions.                                        // 692
                                                                                                      // 693
_.extend(SandstormDb.prototype, {                                                                     // 694
  userGrains: function userGrains (user) {                                                            // 695
    return this.collections.grains.find({ userId: user});                                             // 696
  },                                                                                                  // 697
                                                                                                      // 698
  currentUserGrains: function currentUserGrains () {                                                  // 699
    return this.userGrains(Meteor.userId());                                                          // 700
  },                                                                                                  // 701
                                                                                                      // 702
  getGrain: function getGrain (grainId) {                                                             // 703
    check(grainId, String);                                                                           // 704
    return this.collections.grains.findOne(grainId);                                                  // 705
  },                                                                                                  // 706
                                                                                                      // 707
  userApiTokens: function userApiTokens (user) {                                                      // 708
    return this.collections.apiTokens.find({'owner.user.userId': user});                              // 709
  },                                                                                                  // 710
                                                                                                      // 711
  currentUserApiTokens: function currentUserApiTokens () {                                            // 712
    return this.userApiTokens(Meteor.userId());                                                       // 713
  },                                                                                                  // 714
                                                                                                      // 715
  userActions: function userActions (user) {                                                          // 716
    return this.collections.userActions.find({userId: user});                                         // 717
  },                                                                                                  // 718
                                                                                                      // 719
  currentUserActions: function currentUserActions () {                                                // 720
    return this.userActions(Meteor.userId());                                                         // 721
  },                                                                                                  // 722
                                                                                                      // 723
  getPlan: function (id) {                                                                            // 724
    check(id, String);                                                                                // 725
    var plan = Plans.findOne(id);                                                                     // 726
    if (!plan) {                                                                                      // 727
      throw new Error("no such plan: " + id);                                                         // 728
    }                                                                                                 // 729
    return plan;                                                                                      // 730
  },                                                                                                  // 731
                                                                                                      // 732
  listPlans: function () {                                                                            // 733
    return Plans.find({}, {sort: {price: 1}});                                                        // 734
  },                                                                                                  // 735
                                                                                                      // 736
  getMyPlan: function () {                                                                            // 737
    var user = Meteor.user();                                                                         // 738
    return user && Plans.findOne(user.plan || "free");                                                // 739
  },                                                                                                  // 740
                                                                                                      // 741
  getMyUsage: function (user) {                                                                       // 742
    user = user || Meteor.user();                                                                     // 743
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                              // 744
      if (Meteor.isClient) {                                                                          // 745
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currenly       // 746
        // not reactive.                                                                              // 747
        return user.pseudoUsage;                                                                      // 748
      } else {                                                                                        // 749
        return {                                                                                      // 750
          grains: Grains.find({userId: user._id}).count(),                                            // 751
          storage: user.storageUsage || 0,                                                            // 752
          compute: 0   // not tracked yet                                                             // 753
        };                                                                                            // 754
      }                                                                                               // 755
    } else {                                                                                          // 756
      return {grains: 0, storage: 0, compute: 0};                                                     // 757
    }                                                                                                 // 758
  },                                                                                                  // 759
                                                                                                      // 760
  isUninvitedFreeUser: function () {                                                                  // 761
    if (!Meteor.settings.public.allowUninvited) return false;                                         // 762
                                                                                                      // 763
    var user = Meteor.user();                                                                         // 764
    return user && !user.expires && (!user.plan || user.plan === "free");                             // 765
  },                                                                                                  // 766
                                                                                                      // 767
  getSetting: function (name) {                                                                       // 768
    var setting = Settings.findOne(name);                                                             // 769
    return setting && setting.value;                                                                  // 770
  },                                                                                                  // 771
});                                                                                                   // 772
                                                                                                      // 773
if (Meteor.isServer) {                                                                                // 774
  var Crypto = Npm.require("crypto");                                                                 // 775
  var ContentType = Npm.require("content-type");                                                      // 776
  var Zlib = Npm.require("zlib");                                                                     // 777
                                                                                                      // 778
  var replicaNumber = Meteor.settings.replicaNumber || 0;                                             // 779
                                                                                                      // 780
  var computeStagger = function (n) {                                                                 // 781
    // Compute a fraction in the range [0, 1) such that, for any natural number k, the values         // 782
    // of computeStagger(n) for all n in [1, 2^k) are uniformly distributed between 0 and 1.          // 783
    // The sequence looks like:                                                                       // 784
    //   0, 1/2, 1/4, 3/4, 1/8, 3/8, 5/8, 7/8, 1/16, ...                                              // 785
    //                                                                                                // 786
    // We use this to determine how we'll stagger periodic events performed by this replica.          // 787
    // Notice that this allows us to compute a stagger which is independent of the number of          // 788
    // front-end replicas present; we can add more replicas to the end without affecting how the      // 789
    // earlier ones schedule their events.                                                            // 790
    var denom = 1;                                                                                    // 791
    while (denom <= n) denom <<= 1;                                                                   // 792
    var num = n * 2 - denom + 1;                                                                      // 793
    return num / denom;                                                                               // 794
  }                                                                                                   // 795
                                                                                                      // 796
  var stagger = computeStagger(replicaNumber);                                                        // 797
                                                                                                      // 798
  SandstormDb.periodicCleanup = function (intervalMs, callback) {                                     // 799
    // Register a database cleanup function than should run periodically, roughly once every          // 800
    // interval of the given length.                                                                  // 801
    //                                                                                                // 802
    // In a blackrock deployment with multiple front-ends, the frequency of the cleanup will be       // 803
    // scaled appropriately on the assumption that more data is being generated demanding more        // 804
    // frequent cleanups.                                                                             // 805
                                                                                                      // 806
    check(intervalMs, Number);                                                                        // 807
    check(callback, Function);                                                                        // 808
                                                                                                      // 809
    if (intervalMs < 120000) {                                                                        // 810
      throw new Error("less than 2-minute cleanup interval seems too fast; " +                        // 811
                      "are you using the right units?");                                              // 812
    }                                                                                                 // 813
                                                                                                      // 814
    // Schedule first cleanup to happen at the next intervalMs interval from the epoch, so that       // 815
    // the schedule is independent of the exact startup time.                                         // 816
    var first = intervalMs - Date.now() % intervalMs;                                                 // 817
                                                                                                      // 818
    // Stagger cleanups across replicas so that we don't have all replicas trying to clean the        // 819
    // same data at the same time.                                                                    // 820
    first += Math.floor(intervalMs * computeStagger(replicaNumber));                                  // 821
                                                                                                      // 822
    // If the stagger put us more than an interval away from now, back up.                            // 823
    if (first > intervalMs) first -= intervalMs;                                                      // 824
                                                                                                      // 825
    Meteor.setTimeout(function () {                                                                   // 826
      callback();                                                                                     // 827
      Meteor.setInterval(callback, intervalMs);                                                       // 828
    }, first);                                                                                        // 829
  }                                                                                                   // 830
                                                                                                      // 831
  // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.           // 832
  var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                   // 833
                                                                                                      // 834
  var BufferSmallerThan = function (limit) {                                                          // 835
    return Match.Where(function (buf) {                                                               // 836
      check(buf, Buffer);                                                                             // 837
      return buf.length < limit;                                                                      // 838
    });                                                                                               // 839
  }                                                                                                   // 840
                                                                                                      // 841
  var DatabaseId = Match.Where(function (s) {                                                         // 842
    check(s, String);                                                                                 // 843
    return !!s.match(/^[a-zA-Z0-9_]+$/);                                                              // 844
  });                                                                                                 // 845
                                                                                                      // 846
  addStaticAsset = function (metadata, content) {                                                     // 847
    // Add a new static asset to the database. If `content` is a string rather than a buffer, it      // 848
    // will be automatically gzipped before storage; do not specify metadata.encoding in this case.   // 849
                                                                                                      // 850
    if (typeof content === "string" && !metadata.encoding) {                                          // 851
      content = gzipSync(new Buffer(content, "utf8"));                                                // 852
      metadata.encoding = "gzip";                                                                     // 853
    }                                                                                                 // 854
                                                                                                      // 855
    check(metadata, {                                                                                 // 856
      mimeType: String,                                                                               // 857
      encoding: Match.Optional("gzip")                                                                // 858
    });                                                                                               // 859
    check(content, BufferSmallerThan(1 << 20));                                                       // 860
                                                                                                      // 861
    // Validate content type.                                                                         // 862
    metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                     // 863
                                                                                                      // 864
    var hasher = Crypto.createHash("sha256");                                                         // 865
    hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                       // 866
    hasher.update(content);                                                                           // 867
    var hash = hasher.digest("base64");                                                               // 868
                                                                                                      // 869
    var existing = StaticAssets.findAndModify({                                                       // 870
      query: {hash: hash, refcount: {$gte: 1}},                                                       // 871
      update: {$inc: {refcount: 1}},                                                                  // 872
      fields: {_id: 1, refcount: 1},                                                                  // 873
    });                                                                                               // 874
    if (existing) {                                                                                   // 875
      return existing._id;                                                                            // 876
    }                                                                                                 // 877
                                                                                                      // 878
    return StaticAssets.insert(_.extend({                                                             // 879
      hash: hash,                                                                                     // 880
      content: content,                                                                               // 881
      refcount: 1                                                                                     // 882
    }, metadata));                                                                                    // 883
  }                                                                                                   // 884
                                                                                                      // 885
  SandstormDb.prototype.addStaticAsset = addStaticAsset;                                              // 886
                                                                                                      // 887
  SandstormDb.prototype.refStaticAsset = function (id) {                                              // 888
    // Increment the refcount on an existing static asset.                                            // 889
    //                                                                                                // 890
    // You must call this BEFORE adding the new reference to the DB, in case of failure between       // 891
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big      // 892
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                        // 893
                                                                                                      // 894
    check(id, String);                                                                                // 895
                                                                                                      // 896
    var existing = StaticAssets.findAndModify({                                                       // 897
      query: {hash: hash},                                                                            // 898
      update: {$inc: {refcount: 1}},                                                                  // 899
      fields: {_id: 1, refcount: 1},                                                                  // 900
    });                                                                                               // 901
    if (!existing) {                                                                                  // 902
      throw new Error("refStaticAsset() called on asset that doesn't exist");                         // 903
    }                                                                                                 // 904
  }                                                                                                   // 905
                                                                                                      // 906
  SandstormDb.prototype.unrefStaticAsset = function (id) {                                            // 907
    // Decrement refcount on a static asset and delete if it has reached zero.                        // 908
    //                                                                                                // 909
    // You must call this AFTER removing the reference from the DB, in case of failure between        // 910
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big      // 911
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                        // 912
                                                                                                      // 913
    check(id, String);                                                                                // 914
                                                                                                      // 915
    var existing = StaticAssets.findAndModify({                                                       // 916
      query: {_id: id},                                                                               // 917
      update: {$inc: {refcount: -1}},                                                                 // 918
      fields: {_id: 1, refcount: 1},                                                                  // 919
      new: true,                                                                                      // 920
    });                                                                                               // 921
    if (!existing) {                                                                                  // 922
      console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);        // 923
    } else if (existing.refcount <= 0) {                                                              // 924
      StaticAssets.remove({_id: existing._id});                                                       // 925
    }                                                                                                 // 926
  }                                                                                                   // 927
                                                                                                      // 928
  SandstormDb.prototype.getStaticAsset = function (id) {                                              // 929
    // Get a static asset's mimeType, encoding, and raw content.                                      // 930
                                                                                                      // 931
    check(id, String);                                                                                // 932
                                                                                                      // 933
    var asset = StaticAssets.findOne(id, {fields: {_id: 0, mimeType: 1, encoding: 1, content: 1}});   // 934
    if (asset) {                                                                                      // 935
      // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy       // 936
      //   here.                                                                                      // 937
      asset.content = new Buffer(asset.content);                                                      // 938
    }                                                                                                 // 939
    return asset;                                                                                     // 940
  }                                                                                                   // 941
                                                                                                      // 942
  SandstormDb.prototype.newAssetUpload = function (purpose) {                                         // 943
    check(purpose, {profilePicture: {userId: DatabaseId, identityId: Match.Optional(DatabaseId)}});   // 944
                                                                                                      // 945
    return AssetUploadTokens.insert({                                                                 // 946
      purpose: purpose,                                                                               // 947
      expires: new Date(Date.now() + 300000),  // in 5 minutes                                        // 948
    });                                                                                               // 949
  }                                                                                                   // 950
                                                                                                      // 951
  SandstormDb.prototype.fulfillAssetUpload = function (id) {                                          // 952
    // Indicates that the given asset upload has completed. It will be removed and its purpose        // 953
    // returned. If no matching upload exists, returns undefined.                                     // 954
                                                                                                      // 955
    check(id, String);                                                                                // 956
                                                                                                      // 957
    var upload = AssetUploadTokens.findAndModify({                                                    // 958
      query: {_id: id},                                                                               // 959
      remove: true                                                                                    // 960
    });                                                                                               // 961
                                                                                                      // 962
    if (upload.expires.valueOf() < Date.now()) {                                                      // 963
      return undefined;  // already expired                                                           // 964
    } else {                                                                                          // 965
      return upload.purpose;                                                                          // 966
    }                                                                                                 // 967
  }                                                                                                   // 968
                                                                                                      // 969
  function cleanupExpiredAssetUploads() {                                                             // 970
    AssetUploadTokens.remove({expires: {$lt: Date.now()}});                                           // 971
  }                                                                                                   // 972
                                                                                                      // 973
  // Cleanup tokens every hour.                                                                       // 974
  SandstormDb.periodicCleanup(3600000, cleanupExpiredAssetUploads);                                   // 975
                                                                                                      // 976
  var packageCache = {};                                                                              // 977
  // Package info is immutable. Let's cache to save on mongo queries.                                 // 978
                                                                                                      // 979
  SandstormDb.prototype.getPackage = function (packageId) {                                           // 980
    // Get the given package record. Since package info is immutable, cache the data in the server    // 981
    // to reduce mongo query overhead, since it turns out we have to fetch specific packages a        // 982
    // lot.                                                                                           // 983
                                                                                                      // 984
    if (packageId in packageCache) {                                                                  // 985
      return packageCache[packageId];                                                                 // 986
    }                                                                                                 // 987
                                                                                                      // 988
    var package = Packages.findOne(packageId);                                                        // 989
    if (package && package.status === "ready") {                                                      // 990
      packageCache[packageId] = package;                                                              // 991
    }                                                                                                 // 992
    return package;                                                                                   // 993
  }                                                                                                   // 994
}                                                                                                     // 995
                                                                                                      // 996
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/profile.js                                                                   //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
var makeIdenticon;                                                                                    // 17
var httpProtocol;                                                                                     // 18
var sha256;                                                                                           // 19
                                                                                                      // 20
if (Meteor.isServer) {                                                                                // 21
  Meteor.publish("accountIdentities", function () {                                                   // 22
    if (!Meteor.userId) return [];                                                                    // 23
                                                                                                      // 24
    return [                                                                                          // 25
      Meteor.users.find(Meteor.userId,                                                                // 26
        {fields: {                                                                                    // 27
          "profile":1,                                                                                // 28
          "devName":1,                                                                                // 29
          "expires":1,                                                                                // 30
                                                                                                      // 31
          "services.google.id":1,                                                                     // 32
          "services.google.email":1,                                                                  // 33
          "services.google.verified_email":1,                                                         // 34
          "services.google.name":1,                                                                   // 35
          "services.google.picture":1,                                                                // 36
          "services.google.gender":1,                                                                 // 37
                                                                                                      // 38
          "services.github.id":1,                                                                     // 39
          "services.github.email":1,                                                                  // 40
          "services.github.username":1,                                                               // 41
                                                                                                      // 42
          "services.emailToken.email":1                                                               // 43
        }})                                                                                           // 44
    ];                                                                                                // 45
  });                                                                                                 // 46
                                                                                                      // 47
  makeIdenticon = function (id) {                                                                     // 48
    // We only make identicons client-side.                                                           // 49
    return undefined;                                                                                 // 50
  }                                                                                                   // 51
                                                                                                      // 52
  var Url = Npm.require("url");                                                                       // 53
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                            // 54
                                                                                                      // 55
  var Crypto = Npm.require("crypto");                                                                 // 56
  sha256 = function (data) {                                                                          // 57
    return Crypto.createHash("sha256").update(data).digest("hex");                                    // 58
  };                                                                                                  // 59
} else {                                                                                              // 60
  var identiconCache = {};                                                                            // 61
                                                                                                      // 62
  makeIdenticon = function (id) {                                                                     // 63
    // Given a cryptographic hash as input, generate an identicon. We always use the user's           // 64
    // identity ID as the input hash (the same thing passed to apps as the user ID).                  // 65
                                                                                                      // 66
    // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so        // 67
    // that apps can themselves use identicon.js to produce consistent identicons. As it turns out    // 68
    // identicon.js doesn't use the second half of the hash even if we provide it, but slice it       // 69
    // anyway to be safe.                                                                             // 70
    id = id.slice(0, 32);                                                                             // 71
                                                                                                      // 72
    if (id in identiconCache) {                                                                       // 73
      return identiconCache[id];                                                                      // 74
    }                                                                                                 // 75
                                                                                                      // 76
    // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.      // 77
    var data = new Identicon(id, 512).toString();                                                     // 78
    var result = "data:image/png;base64," + data;                                                     // 79
    identiconCache[id] = result;                                                                      // 80
    return result;                                                                                    // 81
  }                                                                                                   // 82
                                                                                                      // 83
  httpProtocol = window.location.protocol;                                                            // 84
                                                                                                      // 85
  sha256 = function (data) {                                                                          // 86
    return CryptoJS.SHA256(data).toString();                                                          // 87
  };                                                                                                  // 88
}                                                                                                     // 89
                                                                                                      // 90
var GENDERS = {male: "male", female: "female", neutral: "neutral", robot: "robot"};                   // 91
                                                                                                      // 92
function staticAssetUrl(id, staticHost) {                                                             // 93
  if (id) {                                                                                           // 94
    return staticHost + "/" + id;                                                                     // 95
  } else {                                                                                            // 96
    return undefined;                                                                                 // 97
  }                                                                                                   // 98
}                                                                                                     // 99
                                                                                                      // 100
function filterHandle(handle) {                                                                       // 101
  // Convert disallowed letters into underscores, but no more than one underscore in a row.           // 102
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                              // 103
}                                                                                                     // 104
                                                                                                      // 105
function emailToHandle(email) {                                                                       // 106
  // Turn an email address into a handle.                                                             // 107
                                                                                                      // 108
  if (!email) return undefined;                                                                       // 109
                                                                                                      // 110
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters        // 111
  // on the same account.                                                                             // 112
  var parts = email.split("@");                                                                       // 113
  var base = filterHandle(parts[0].split("+")[0]);                                                    // 114
                                                                                                      // 115
  var domain = (parts[1]||"").split(".");                                                             // 116
  if (domain[domain.length - 1] === "name") {                                                         // 117
    // Oh, a .name domain. Let's use                                                                  // 118
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                // 119
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info",           // 120
                         "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"],          // 121
                         base)) {                                                                     // 122
    // This is probably an address at a vanity domain. Use the domain itself as the handle.           // 123
    base = filterHandle(domain[0]);                                                                   // 124
  }                                                                                                   // 125
                                                                                                      // 126
  return filterHandle(base);                                                                          // 127
}                                                                                                     // 128
                                                                                                      // 129
function identityId(service, id) {                                                                    // 130
  return sha256(service + ":" + id);                                                                  // 131
}                                                                                                     // 132
                                                                                                      // 133
function googleIdentity(user, staticHost) {                                                           // 134
  var google = (user.services && user.services.google) || {};                                         // 135
  var profile = user.profile || {};                                                                   // 136
  var id = identityId("google", google.id);                                                           // 137
                                                                                                      // 138
  return {                                                                                            // 139
    service: "google",                                                                                // 140
    id: id,                                                                                           // 141
    name: profile.name || google.name || "Name Unknown",                                              // 142
    email: profile.email || google.email,                                                             // 143
    handle: profile.handle || emailToHandle(google.email) ||                                          // 144
            filterHandle(profile.name || google.name) || "unknown",                                   // 145
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 146
    pronoun: profile.pronoun || GENDERS[google.gender] || "neutral",                                  // 147
  }                                                                                                   // 148
}                                                                                                     // 149
                                                                                                      // 150
function githubIdentity(user, staticHost) {                                                           // 151
  var github = (user.services && user.services.github) || {};                                         // 152
  var profile = user.profile || {};                                                                   // 153
  var id = identityId("github", github.id);                                                           // 154
                                                                                                      // 155
  return {                                                                                            // 156
    service: "github",                                                                                // 157
    id: id,                                                                                           // 158
    name: profile.name || github.username || "Name Unknown",                                          // 159
    email: profile.email || github.email,                                                             // 160
    handle: profile.handle || filterHandle(github.username) ||                                        // 161
            filterHandle(profile.name) || "unknown",                                                  // 162
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 163
    pronoun: profile.pronoun,                                                                         // 164
  }                                                                                                   // 165
}                                                                                                     // 166
                                                                                                      // 167
function emailIdentity(user, staticHost) {                                                            // 168
  var email = (user.services && user.services.emailToken) || {};                                      // 169
  var profile = user.profile || {};                                                                   // 170
  var id = identityId("email", email.email);                                                          // 171
                                                                                                      // 172
  return {                                                                                            // 173
    service: "email",                                                                                 // 174
    id: id,                                                                                           // 175
    name: profile.name || email.email.split("@")[0] || "Name Unknown",                                // 176
    email: profile.email || email.email,                                                              // 177
    handle: profile.handle || emailToHandle(email.email) ||                                           // 178
            filterHandle(profile.name) || "unknown",                                                  // 179
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 180
    pronoun: profile.pronoun,                                                                         // 181
  }                                                                                                   // 182
}                                                                                                     // 183
                                                                                                      // 184
function devIdentity(user, staticHost) {                                                              // 185
  var email = (user.services && user.services.emailToken) || {};                                      // 186
  var profile = user.profile || {};                                                                   // 187
  var name = user.devName.split(" ")[0].toLowerCase();                                                // 188
  var id = identityId("dev", user.devName);                                                           // 189
                                                                                                      // 190
  return {                                                                                            // 191
    service: "dev",                                                                                   // 192
    id: id,                                                                                           // 193
    name: profile.name || user.devName,                                                               // 194
    handle: profile.handle || filterHandle(name),                                                     // 195
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 196
    pronoun: profile.pronoun ||                                                                       // 197
             (_.contains(["alice", "carol", "eve"], name) ? "female" :                                // 198
              _.contains(["bob", "dave"], name) ? "male" : "neutral"),                                // 199
  }                                                                                                   // 200
}                                                                                                     // 201
                                                                                                      // 202
function demoIdentity(user, staticHost) {                                                             // 203
  var email = (user.services && user.services.emailToken) || {};                                      // 204
  var profile = user.profile || {};                                                                   // 205
  var id = identityId("demo", user._id);                                                              // 206
                                                                                                      // 207
  return {                                                                                            // 208
    service: "demo",                                                                                  // 209
    id: id,                                                                                           // 210
    name: profile.name || "Demo User",                                                                // 211
    handle: profile.handle || "demo",                                                                 // 212
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 213
    pronoun: profile.pronoun || "neutral",                                                            // 214
  }                                                                                                   // 215
}                                                                                                     // 216
                                                                                                      // 217
SandstormDb.getVerifiedEmails = function (user) {                                                     // 218
  var result = [];                                                                                    // 219
                                                                                                      // 220
  var services = user.services || {};                                                                 // 221
  var google = services.google || {};                                                                 // 222
  if (google.email && google.verified_email) result.push(google.email);                               // 223
  if (services.emailToken) result.push(services.emailToken.email);                                    // 224
                                                                                                      // 225
  // TODO(soon): Verification of email addresses -- perhaps through asking the user to log in as      // 226
  //   the given identity?                                                                            // 227
                                                                                                      // 228
  return result;                                                                                      // 229
}                                                                                                     // 230
                                                                                                      // 231
SandstormDb.getUserIdentities = function (user) {                                                     // 232
  // Given a user object, return all of the user's identities.                                        // 233
  //                                                                                                  // 234
  // On the client, must be subscribed "accountIdentities" for the user.                              // 235
                                                                                                      // 236
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                  // 237
                                                                                                      // 238
  var result = [];                                                                                    // 239
  if (user && user.services) {                                                                        // 240
    if ("google" in user.services) {                                                                  // 241
      result.push(googleIdentity(user, staticHost));                                                  // 242
    }                                                                                                 // 243
    if ("github" in user.services) {                                                                  // 244
      result.push(githubIdentity(user, staticHost));                                                  // 245
    }                                                                                                 // 246
    if ("emailToken" in user.services) {                                                              // 247
      result.push(emailIdentity(user, staticHost));                                                   // 248
    }                                                                                                 // 249
    if ("devName" in user) {                                                                          // 250
      result.push(devIdentity(user, staticHost));                                                     // 251
    }                                                                                                 // 252
    if ("expires" in user) {                                                                          // 253
      result.push(demoIdentity(user, staticHost));                                                    // 254
    }                                                                                                 // 255
  }                                                                                                   // 256
  return result;                                                                                      // 257
}                                                                                                     // 258
                                                                                                      // 259
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/user.js                                                                      //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
var Future = Npm.require("fibers/future");                                                            // 17
                                                                                                      // 18
userPictureUrl = function (user) {                                                                    // 19
  if (user.services && !(user.profile && user.profile.picture)) {                                     // 20
    // Try to determine user's avatar URL from login service.                                         // 21
                                                                                                      // 22
    var google = user.services.google;                                                                // 23
    if (google && google.picture) {                                                                   // 24
      return google.picture;                                                                          // 25
    }                                                                                                 // 26
                                                                                                      // 27
    var github = user.services.github;                                                                // 28
    if (github && github.id) {                                                                        // 29
      return "https://avatars.githubusercontent.com/u/" + github.id;                                  // 30
    }                                                                                                 // 31
                                                                                                      // 32
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be     // 33
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and         // 34
    // Github are different because they are actually the identity providers, so they already know    // 35
    // the user logged in.                                                                            // 36
  }                                                                                                   // 37
}                                                                                                     // 38
                                                                                                      // 39
fetchPicture = function (url) {                                                                       // 40
  try {                                                                                               // 41
    var result = HTTP.get(url, {                                                                      // 42
      npmRequestOptions: { encoding: null },                                                          // 43
      timeout: 5000                                                                                   // 44
    });                                                                                               // 45
                                                                                                      // 46
    var metadata = {};                                                                                // 47
                                                                                                      // 48
    metadata.mimeType = result.headers["content-type"];                                               // 49
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 &&                                       // 50
        metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {                                      // 51
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                 // 52
    }                                                                                                 // 53
    var enc = result.headers["content-encoding"];                                                     // 54
    if (enc && enc !== "identity") {                                                                  // 55
      metadata.encoding = enc;                                                                        // 56
    }                                                                                                 // 57
                                                                                                      // 58
    return addStaticAsset(metadata, result.content);                                                  // 59
  } catch (err) {                                                                                     // 60
    console.error("failed to fetch user profile picture:", url, err.stack);                           // 61
  }                                                                                                   // 62
}                                                                                                     // 63
                                                                                                      // 64
var ValidHandle = Match.Where(function (handle) {                                                     // 65
  check(handle, String);                                                                              // 66
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                        // 67
});                                                                                                   // 68
                                                                                                      // 69
Accounts.onCreateUser(function (options, user) {                                                      // 70
  // The first non-dev user to sign in should be automatically upgraded to admin.                     // 71
  // Dev users are identified by having the devName field.                                            // 72
  if (Meteor.users.find({devName: {$exists: 0}}).count() === 0 && !user.devName) {                    // 73
    user.isAdmin = true;                                                                              // 74
    user.signupKey = "admin";                                                                         // 75
  }                                                                                                   // 76
                                                                                                      // 77
  // Check profile.                                                                                   // 78
  if (options.profile) {                                                                              // 79
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                              // 80
    check(options.profile, Match.ObjectIncluding({                                                    // 81
      name: Match.OneOf(null, Match.Optional(String)),                                                // 82
      handle: Match.Optional(ValidHandle),                                                            // 83
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot")),                     // 84
    }));                                                                                              // 85
                                                                                                      // 86
    user.profile = options.profile;                                                                   // 87
  } else {                                                                                            // 88
    user.profile = {};                                                                                // 89
  }                                                                                                   // 90
                                                                                                      // 91
  // Try downloading avatar.                                                                          // 92
  var url = userPictureUrl(user);                                                                     // 93
  if (url) {                                                                                          // 94
    var assetId = fetchPicture(url);                                                                  // 95
    if (assetId) {                                                                                    // 96
      user.profile.picture = assetId;                                                                 // 97
    }                                                                                                 // 98
  }                                                                                                   // 99
                                                                                                      // 100
  return user;                                                                                        // 101
});                                                                                                   // 102
                                                                                                      // 103
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/migrations.js                                                                //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                   // 1
//                                                                                                    // 2
// Because migrations can experience partial failure and likely have                                  // 3
// side-effects, we should be careful to make sure all migrations are                                 // 4
// idempotent and safe to accidentally run multiple times.                                            // 5
                                                                                                      // 6
var Future = Npm.require("fibers/future");                                                            // 7
                                                                                                      // 8
var updateLoginStyleToRedirect = function() {                                                         // 9
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;          // 10
  ["google", "github"].forEach(function(serviceName) {                                                // 11
    var config = configurations.findOne({service: serviceName});                                      // 12
    if (config && config.loginStyle !== "redirect") {                                                 // 13
      configurations.update({service: serviceName}, {$set: {loginStyle: "redirect"}});                // 14
    }                                                                                                 // 15
  });                                                                                                 // 16
};                                                                                                    // 17
                                                                                                      // 18
var enableLegacyOAuthProvidersIfNotInSettings = function() {                                          // 19
  // In the before time, Google and Github login were enabled by default.                             // 20
  //                                                                                                  // 21
  // This actually didn't make much sense, required the first user to configure                       // 22
  // OAuth, and had some trust-the-first-user properties that weren't totally                         // 23
  // secure.                                                                                          // 24
  //                                                                                                  // 25
  // Now, we have admin-token, so we wish to disable all logins by default                            // 26
  // (since they need to be configured anyway) but since users may have never                         // 27
  // explicitly told Sandstorm that Google or Github login should be enabled,                         // 28
  // we can't just use the value in the Settings collection, since it might                           // 29
  // never have been set.                                                                             // 30
  //                                                                                                  // 31
  // Thus, if the service is configured but absent from Settings, we should                           // 32
  // explicitly enable it in Settings, and then the rest of the logic can just                        // 33
  // depend on what value is in Settings and default to false without breaking                        // 34
  // user installations.                                                                              // 35
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;          // 36
  ["google", "github"].forEach(function(serviceName) {                                                // 37
    var config = configurations.findOne({service: serviceName});                                      // 38
    var serviceConfig = Settings.findOne({_id: serviceName});                                         // 39
    if (config && !serviceConfig) {                                                                   // 40
      // Only explicitly enable the login service if:                                                 // 41
      // 1) the service is already configured                                                         // 42
      // 2) there is no sandstorm configuration already present (the user was                         // 43
      //    using the previous default behavior).                                                     // 44
      Settings.insert({_id: serviceName, value: true});                                               // 45
    }                                                                                                 // 46
  });                                                                                                 // 47
};                                                                                                    // 48
                                                                                                      // 49
var denormalizeInviteInfo = function() {                                                              // 50
  // When a user is invited via a signup token, the `signupKey` field of their user table entry       // 51
  // has always been populated to indicate the key they used. This points into the SignupKeys table   // 52
  // which has more information about the key, namely a freeform note entered by the admin when       // 53
  // they created the key. In the case that the email invite form was used, the note has the form     // 54
  // "E-mail invite to <address>".                                                                    // 55
  //                                                                                                  // 56
  // Later, we decided it was useful to indicate in the users table visible to the admin              // 57
  // information about the invite terms. Namely, for email invites we want to show the address        // 58
  // and for others we want to show the note. To make this efficient, fields `signupNote` and         // 59
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing       // 60
  // from the SignupKeys table.                                                                       // 61
                                                                                                      // 62
  Meteor.users.find().forEach(function (user) {                                                       // 63
    if (user.signupKey && (typeof user.signupKey) === "string" && user.signupKey !== "admin") {       // 64
      var signupInfo = SignupKeys.findOne(user.signupKey);                                            // 65
      if (signupInfo && signupInfo.note) {                                                            // 66
        var newFields = { signupNote: signupInfo.note };                                              // 67
                                                                                                      // 68
        var prefix = "E-mail invite to ";                                                             // 69
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                              // 70
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                               // 71
        }                                                                                             // 72
                                                                                                      // 73
        Meteor.users.update(user._id, {$set: newFields});                                             // 74
      }                                                                                               // 75
    }                                                                                                 // 76
  });                                                                                                 // 77
}                                                                                                     // 78
                                                                                                      // 79
function mergeRoleAssignmentsIntoApiTokens() {                                                        // 80
  RoleAssignments.find().forEach(function(roleAssignment) {                                           // 81
    ApiTokens.insert({                                                                                // 82
      grainId: roleAssignment.grainId,                                                                // 83
      userId: roleAssignment.sharer,                                                                  // 84
      roleAssignment: roleAssignment.roleAssignment,                                                  // 85
      petname: roleAssignment.petname,                                                                // 86
      created: roleAssignment.created,                                                                // 87
      owner: {user: {userId: roleAssignment.recipient,                                                // 88
                     title: roleAssignment.title}},                                                   // 89
    });                                                                                               // 90
  });                                                                                                 // 91
}                                                                                                     // 92
                                                                                                      // 93
function fixOasisStorageUsageStats() {                                                                // 94
  // Before storage quota enforcement was first implemented, there were some bugs with the way the    // 95
  // Oasis backend tracked storage usage, which were not initially noticed since the resulting        // 96
  // number wasn't acutally used for anything.                                                        // 97
  //                                                                                                  // 98
  // With the underlying bugs fixed, simply starting up each grain (supervisor only, not the app)     // 99
  // and then immediately shutting it down should cause the sizes to be recalculated correctly.       // 100
  //                                                                                                  // 101
  // This problem only applies to Oasis.                                                              // 102
                                                                                                      // 103
  if (Meteor.settings.public.quotaEnabled) {                                                          // 104
    var dummyCommand = {                                                                              // 105
      argv: ["foo", "bar"],                                                                           // 106
      environ: []                                                                                     // 107
    };                                                                                                // 108
                                                                                                      // 109
    Grains.find().forEach(function (grain) {                                                          // 110
      try {                                                                                           // 111
        console.log("refreshing grain:", grain._id);                                                  // 112
                                                                                                      // 113
        // We take a highly synchronous approach here so that the logging is understandable.          // 114
        var supervisor = waitPromise(sandstormBackend.startGrain(                                     // 115
            grain.userId, grain._id, grain.packageId, dummyCommand, false, false)).supervisor;        // 116
        try {                                                                                         // 117
          waitPromise(supervisor.shutdown());                                                         // 118
          throw new Error("shutdown() didn't disconnect!");                                           // 119
        } catch (err) {                                                                               // 120
          if (err.kjType !== "disconnected") throw err;                                               // 121
        }                                                                                             // 122
      } catch (err) {                                                                                 // 123
        console.log(err.stack);                                                                       // 124
      }                                                                                               // 125
    });                                                                                               // 126
  }                                                                                                   // 127
}                                                                                                     // 128
                                                                                                      // 129
function fetchProfilePictures() {                                                                     // 130
  Meteor.users.find({}).forEach(function (user) {                                                     // 131
    var url = userPictureUrl(user);                                                                   // 132
    if (url) {                                                                                        // 133
      console.log("Fetching user picture:", url);                                                     // 134
      var assetId = fetchPicture(url);                                                                // 135
      if (assetId) {                                                                                  // 136
        Meteor.users.update(user._id, {$set: {"profile.picture": assetId}});                          // 137
      }                                                                                               // 138
    }                                                                                                 // 139
  });                                                                                                 // 140
}                                                                                                     // 141
                                                                                                      // 142
function assignPlans() {                                                                              // 143
  if (Meteor.settings.public.quotaEnabled && SandstormDb.paymentsMigrationHook) {                     // 144
    SandstormDb.paymentsMigrationHook(SignupKeys, Plans.find().fetch());                              // 145
  }                                                                                                   // 146
}                                                                                                     // 147
                                                                                                      // 148
function removeKeyrings() {                                                                           // 149
  // These blobs full of public keys were not intended to find their way into mongo and while         // 150
  // harmless they slow things down because they're huge. Remove them.                                // 151
  Packages.update({"manifest.metadata.pgpKeyring": {$exists: true}},                                  // 152
      {$unset: {"manifest.metadata.pgpKeyring": ""}},                                                 // 153
      {multi: true});                                                                                 // 154
}                                                                                                     // 155
                                                                                                      // 156
function useLocalizedTextInUserActions() {                                                            // 157
  function toLocalizedText(newObj, oldObj, field) {                                                   // 158
    if (field in oldObj) {                                                                            // 159
      if (typeof oldObj[field] === "string") {                                                        // 160
        newObj[field] = {defaultText: oldObj[field]};                                                 // 161
      } else {                                                                                        // 162
        newObj[field] = oldObj[field];                                                                // 163
      }                                                                                               // 164
    }                                                                                                 // 165
  }                                                                                                   // 166
  UserActions.find({}).forEach(function (userAction) {                                                // 167
    var fields = {};                                                                                  // 168
    toLocalizedText(fields, userAction, "appTitle");                                                  // 169
    toLocalizedText(fields, userAction, "title");                                                     // 170
    toLocalizedText(fields, userAction, "nounPhrase");                                                // 171
    UserActions.update(userAction._id, {$set: fields});                                               // 172
  });                                                                                                 // 173
}                                                                                                     // 174
                                                                                                      // 175
// This must come after all the functions named within are defined.                                   // 176
// Only append to this list!  Do not modify or remove list entries;                                   // 177
// doing so is likely change the meaning and semantics of user databases.                             // 178
var MIGRATIONS = [                                                                                    // 179
  updateLoginStyleToRedirect,                                                                         // 180
  enableLegacyOAuthProvidersIfNotInSettings,                                                          // 181
  denormalizeInviteInfo,                                                                              // 182
  mergeRoleAssignmentsIntoApiTokens,                                                                  // 183
  fixOasisStorageUsageStats,                                                                          // 184
  fetchProfilePictures,                                                                               // 185
  assignPlans,                                                                                        // 186
  removeKeyrings,                                                                                     // 187
  useLocalizedTextInUserActions,                                                                      // 188
];                                                                                                    // 189
                                                                                                      // 190
function migrateToLatest() {                                                                          // 191
  if (Meteor.settings.replicaNumber) {                                                                // 192
    // This is a replica. Wait for the first replica to perform migrations.                           // 193
                                                                                                      // 194
    console.log("Waiting for migrations on replica zero...");                                         // 195
                                                                                                      // 196
    var done = new Future();                                                                          // 197
    var change = function (doc) {                                                                     // 198
      console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);            // 199
      if (doc.value >= MIGRATIONS.length) done.return();                                              // 200
    }                                                                                                 // 201
    var observer = Migrations.find({_id: "migrations_applied"}).observe({                             // 202
      added: change,                                                                                  // 203
      changed: change                                                                                 // 204
    });                                                                                               // 205
                                                                                                      // 206
    done.wait();                                                                                      // 207
    observer.stop();                                                                                  // 208
    console.log("Migrations have completed on replica zero.");                                        // 209
                                                                                                      // 210
  } else {                                                                                            // 211
    var applied = Migrations.findOne({_id: "migrations_applied"});                                    // 212
    var start;                                                                                        // 213
    if (!applied) {                                                                                   // 214
      // Migrations table is not yet seeded with a value.  This means it has                          // 215
      // applied 0 migrations.  Persist this.                                                         // 216
      Migrations.insert({_id: "migrations_applied", value: 0});                                       // 217
      start = 0;                                                                                      // 218
    } else {                                                                                          // 219
      start = applied.value;                                                                          // 220
    }                                                                                                 // 221
    console.log("Migrations already applied: " + start + "/" + MIGRATIONS.length);                    // 222
                                                                                                      // 223
    for (var i = start ; i < MIGRATIONS.length ; i++) {                                               // 224
      // Apply migration i, then record that migration i was successfully run.                        // 225
      console.log("Applying migration " + (i+1));                                                     // 226
      MIGRATIONS[i]();                                                                                // 227
      Migrations.update({_id: "migrations_applied"}, {$set: {value: i+1}});                           // 228
      console.log("Applied migration " + (i+1));                                                      // 229
    }                                                                                                 // 230
  }                                                                                                   // 231
}                                                                                                     // 232
                                                                                                      // 233
// Apply all migrations on startup.                                                                   // 234
Meteor.startup(migrateToLatest);                                                                      // 235
                                                                                                      // 236
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
